import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;

class GenJava implements Visitor<Void>{
  ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES | ClassWriter.COMPUTE_MAXS);
  MethodVisitor mv = null;  

  Map<String,Integer> env = new HashMap<>();
  Map<String,Tree.FunDef> funs = new HashMap<>(); 

  
  @Override public Void visit(Tree.FunDef fd){
    var args = "";
    for (var a:fd.args) args+="I";
    mv = cw.visitMethod(Opcodes.ACC_PUBLIC | Opcodes.ACC_STATIC, fd.name, "("+args+")I", null, null);
    var i = 0;
    for (var a:fd.args)env.put(a,i++);
    fd.body.welcome(this);
    mv.visitInsn(Opcodes.IRETURN);
    mv.visitMaxs(1, 1);    
    mv.visitEnd();
    return null;
  }
  
  @Override public Void visit(Tree.Prog prog){
    for (var fun:prog.funs) funs.put(fun.name,fun);

    cw.visit(Opcodes.V1_8, Opcodes.ACC_PUBLIC + Opcodes.ACC_SUPER, "Module", null, "java/lang/Object", null);
    cw.visitSource("Module.java", null);

    for(var fundef:prog.funs) fundef.welcome(this);

    mv = cw.visitMethod(Opcodes.ACC_PUBLIC | Opcodes.ACC_STATIC, "main", "([Ljava/lang/String;)V", null, null);
    mv.visitFieldInsn(Opcodes.GETSTATIC, "java/lang/System", "out", "Ljava/io/PrintStream;");

    prog.expr.welcome(this);

    mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/io/PrintStream", "println", "(I)V", false);
    mv.visitInsn(Opcodes.RETURN);
    mv.visitMaxs(1, 1);
    
    mv.visitEnd();

    cw.visitEnd();

    try {
      OutputStream out = new FileOutputStream("Module.class");
      out.write(cw.toByteArray());
      out.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    }
    return null;
  }
  @Override public Void visit(Tree.MultExpr e){
    e.op1.welcome(this);
    e.op2.welcome(this);
    mv.visitInsn(Opcodes.IMUL);
    return null;
  }
  @Override public Void visit(Tree.AddExpr e){
    e.op1.welcome(this);
    e.op2.welcome(this);
    mv.visitInsn(Opcodes.IADD);

    return null;
  }

  @Override public Void visit(Tree.IntLit e){
    mv.visitLdcInsn(e.n);
    return null;
  }

  public static void main(String[] args) throws Exception{
    LParser parser = new LParser(System.in);
    var gj = new GenJava();
    System.out.println(parser.program().welcome(gj));

  }

@Override
public Void visit(Tree.Var e) {
	mv.visitVarInsn(Opcodes.ILOAD, env.get(e.name));
	return null;
}

@Override
public Void visit(Tree.FunCall e) {
	e.args.forEach(arg -> arg.welcome(this));
    var fd = funs.get(e.name);
    var args="";
	for (var a:fd.args) args+="I";
	mv.visitMethodInsn(Opcodes.INVOKESTATIC,"Module", e.name, "("+args+")I", false);
	return null;
}
}
